from typing import Dict, Optional, cast

FFECT_RANDOM = "random"
EFFECT_COLORLOOP = "colorloop"
EFFECT_RED_FADE = "red_fade"
EFFECT_GREEN_FADE = "green_fade"
EFFECT_BLUE_FADE = "blue_fade"
EFFECT_YELLOW_FADE = "yellow_fade"
EFFECT_CYAN_FADE = "cyan_fade"
EFFECT_PURPLE_FADE = "purple_fade"
EFFECT_WHITE_FADE = "white_fade"
EFFECT_RED_GREEN_CROSS_FADE = "rg_cross_fade"
EFFECT_RED_BLUE_CROSS_FADE = "rb_cross_fade"
EFFECT_GREEN_BLUE_CROSS_FADE = "gb_cross_fade"
EFFECT_COLORSTROBE = "colorstrobe"
EFFECT_RED_STROBE = "red_strobe"
EFFECT_GREEN_STROBE = "green_strobe"
EFFECT_BLUE_STROBE = "blue_strobe"
EFFECT_YELLOW_STROBE = "yellow_strobe"
EFFECT_CYAN_STROBE = "cyan_strobe"
EFFECT_PURPLE_STROBE = "purple_strobe"
EFFECT_WHITE_STROBE = "white_strobe"
EFFECT_COLORJUMP = "colorjump"
EFFECT_CUSTOM = "custom"

EFFECT_MAP = {
    EFFECT_COLORLOOP: 0x25,
    EFFECT_RED_FADE: 0x26,
    EFFECT_GREEN_FADE: 0x27,
    EFFECT_BLUE_FADE: 0x28,
    EFFECT_YELLOW_FADE: 0x29,
    EFFECT_CYAN_FADE: 0x2A,
    EFFECT_PURPLE_FADE: 0x2B,
    EFFECT_WHITE_FADE: 0x2C,
    EFFECT_RED_GREEN_CROSS_FADE: 0x2D,
    EFFECT_RED_BLUE_CROSS_FADE: 0x2E,
    EFFECT_GREEN_BLUE_CROSS_FADE: 0x2F,
    EFFECT_COLORSTROBE: 0x30,
    EFFECT_RED_STROBE: 0x31,
    EFFECT_GREEN_STROBE: 0x32,
    EFFECT_BLUE_STROBE: 0x33,
    EFFECT_YELLOW_STROBE: 0x34,
    EFFECT_CYAN_STROBE: 0x35,
    EFFECT_PURPLE_STROBE: 0x36,
    EFFECT_WHITE_STROBE: 0x37,
    EFFECT_COLORJUMP: 0x38,
}

EFFECT_ID_NAME = {v: k for k, v in EFFECT_MAP.items()}
EFFECT_CUSTOM_CODE = 0x60

EFFECT_LIST = sorted(EFFECT_MAP)

ADDRESSABLE_EFFECT_ID_NAME = {
    1: "RBM 1",
    2: "RBM 2",
    3: "RBM 3",
    4: "RBM 4",
    5: "RBM 5",
    6: "RBM 6",
    7: "RBM 7",
    8: "RBM 8",
    9: "RBM 9",
    10: "RBM 10",
    11: "RBM 11",
    12: "RBM 12",
    13: "RBM 13",
    14: "RBM 14",
    15: "RBM 15",
    16: "RBM 16",
    17: "RBM 17",
    18: "RBM 18",
    19: "RBM 19",
    20: "RBM 20",
    21: "RBM 21",
    22: "RBM 22",
    23: "RBM 23",
    24: "RBM 24",
    25: "RBM 25",
    26: "RBM 26",
    27: "RBM 27",
    28: "RBM 28",
    29: "RBM 29",
    30: "RBM 30",
    31: "RBM 31",
    32: "RBM 32",
    33: "RBM 33",
    34: "RBM 34",
    35: "RBM 35",
    36: "RBM 36",
    37: "RBM 37",
    38: "RBM 38",
    39: "RBM 39",
    40: "RBM 40",
    41: "RBM 41",
    42: "RBM 42",
    43: "RBM 43",
    44: "RBM 44",
    45: "RBM 45",
    46: "RBM 46",
    47: "RBM 47",
    48: "RBM 48",
    49: "RBM 49",
    50: "RBM 50",
    51: "RBM 51",
    52: "RBM 52",
    53: "RBM 53",
    54: "RBM 54",
    55: "RBM 55",
    56: "RBM 56",
    57: "RBM 57",
    58: "RBM 58",
    59: "RBM 59",
    60: "RBM 60",
    61: "RBM 61",
    62: "RBM 62",
    63: "RBM 63",
    64: "RBM 64",
    65: "RBM 65",
    66: "RBM 66",
    67: "RBM 67",
    68: "RBM 68",
    69: "RBM 69",
    70: "RBM 70",
    71: "RBM 71",
    72: "RBM 72",
    73: "RBM 73",
    74: "RBM 74",
    75: "RBM 75",
    76: "RBM 76",
    77: "RBM 77",
    78: "RBM 78",
    79: "RBM 79",
    80: "RBM 80",
    81: "RBM 81",
    82: "RBM 82",
    83: "RBM 83",
    84: "RBM 84",
    85: "RBM 85",
    86: "RBM 86",
    87: "RBM 87",
    88: "RBM 88",
    89: "RBM 89",
    90: "RBM 90",
    91: "RBM 91",
    92: "RBM 92",
    93: "RBM 93",
    94: "RBM 94",
    95: "RBM 95",
    96: "RBM 96",
    97: "RBM 97",
    98: "RBM 98",
    99: "RBM 99",
    100: "RBM 100",
    101: "RBM 101",  # Not in the Magic Home App (only set by remote)
    102: "RBM 102",  # Not in the Magic Home App (only set by remote)
    255: "Circulate all modes",  # Cycles all
}
ADDRESSABLE_EFFECT_NAME_ID = {v: k for k, v in ADDRESSABLE_EFFECT_ID_NAME.items()}

ASSESSABLE_MULTI_COLOR_ID_NAME = {
    1: "Multi Color Static",
    2: "Multi Color Running Water",
    3: "Multi Color Strobe",
    4: "Multi Color Jump",
    5: "Multi Color Breathing",
}

ASSESSABLE_MULTI_COLOR_NAME_ID = {
    v: k for k, v in ASSESSABLE_MULTI_COLOR_ID_NAME.items()
}

ORIGINAL_ADDRESSABLE_EFFECT_ID_NAME = {
    1: "Circulate all modes",
    2: "7 colors change gradually",
    3: "7 colors run in olivary",
    4: "7 colors change quickly",
    5: "7 colors strobe-flash",
    6: "7 colors running, 1 point from start to end and return back",
    7: "7 colors running, multi points from start to end and return back",
    8: "7 colors overlay, multi points from start to end and return back",
    9: "7 colors overlay, multi points from the middle to the both ends and return back",
    10: "7 colors flow gradually, from start to end and return back",
    11: "Fading out run, 7 colors from start to end and return back",
    12: "Runs in olivary, 7 colors from start to end and return back",
    13: "Fading out run, 7 colors start with white color from start to end and return back",
    14: "Run circularly, 7 colors with black background, 1point from start to end",
    15: "Run circularly, 7 colors with red background, 1point from start to end",
    16: "Run circularly, 7 colors with green background, 1point from start to end",
    17: "Run circularly, 7 colors with blue background, 1point from start to end",
    18: "Run circularly, 7 colors with yellow background, 1point from start to end",
    19: "Run circularly, 7 colors with purple background, 1point from start to end",
    20: "Run circularly, 7 colors with cyan background, 1point from start to end",
    21: "Run circularly, 7 colors with white background, 1point from start to end",
    22: "Run circularly, 7 colors with black background, 1point from end to start",
    23: "Run circularly, 7 colors with red background, 1point from end to start",
    24: "Run circularly, 7 colors with green background, 1point from end to start",
    25: "Run circularly, 7 colors with blue background, 1point from end to start",
    26: "Run circularly, 7 colors with yellow background, 1point from end to start",
    27: "Run circularly, 7 colors with purple background, 1point from end to start",
    28: "Run circularly, 7 colors with cyan background, 1point from end to start",
    29: "Run circularly, 7 colors with white background, 1point from end to start",
    30: "Run circularly, 7 colors with black background, 1point from start to end and return back",
    31: "Run circularly, 7 colors with red background, 1point from start to end and return back",
    32: "Run circularly, 7 colors with green background, 1point from start to end and return back",
    33: "Run circularly, 7 colors with blue background, 1point from start to end and return back",
    34: "Run circularly, 7 colors with yellow background, 1point from start to end and return back",
    35: "Run circularly, 7 colors with purple background, 1point from start to end and return back",
    36: "Run circularly, 7 colors with cyan background, 1point from start to end and return back",
    37: "Run circularly, 7 colors with white background, 1point from start to end and return back",
    38: "Run circularly, 7 colors with black background, 1point from middle to both ends",
    39: "Run circularly, 7 colors with red background, 1point from middle to both ends",
    40: "Run circularly, 7 colors with green background, 1point from middle to both ends",
    41: "Run circularly, 7 colors with blue background, 1point from middle to both ends",
    42: "Run circularly, 7 colors with yellow background, 1point from middle to both ends",
    43: "Run circularly, 7 colors with purple background, 1point from middle to both ends",
    44: "Run circularly, 7 colors with cyan background, 1point from middle to both ends",
    45: "Run circularly, 7 colors with white background, 1point from middle to both ends",
    46: "Run circularly, 7 colors with black background, 1point from both ends to middle",
    47: "Run circularly, 7 colors with red background, 1point from both ends to middle",
    48: "Run circularly, 7 colors with green background, 1point from both ends to middle",
    49: "Run circularly, 7 colors with blue background, 1point from both ends to middle",
    50: "Run circularly, 7 colors with yellow background, 1point from both ends to middle",
    51: "Run circularly, 7 colors with purple background, 1point from both ends to middle",
    52: "Run circularly, 7 colors with cyan background, 1point from both ends to middle",
    53: "Run circularly, 7 colors with white background, 1point from both ends to middle",
    54: "Run circularly, 7 colors with black background, 1point from middle to both ends and return back",
    55: "Run circularly, 7 colors with red background, 1point from middle to both ends and return back",
    56: "Run circularly, 7 colors with green background, 1point from middle to both ends and return back",
    57: "Run circularly, 7 colors with blue background, 1point from middle to both ends and return back",
    58: "Run circularly, 7 colors with yellow background, 1point from middle to both ends and return back",
    59: "Run circularly, 7 colors with purple background, 1point from middle to both ends and return back",
    60: "Run circularly, 7 colors with cyan background, 1point from middle to both ends and return back",
    61: "Run circularly, 7 colors with white background, 1point from middle to both ends and return back",
    62: "Overlay circularly, 7 colors with black background from start to end",
    63: "Overlay circularly, 7 colors with red background from start to end",
    64: "Overlay circularly, 7 colors with green background from start to end",
    65: "Overlay circularly, 7 colors with blue background from start to end",
    66: "Overlay circularly, 7 colors with yellow background from start to end",
    67: "Overlay circularly, 7 colors with purple background from start to end",
    68: "Overlay circularly, 7 colors with cyan background from start to end",
    69: "Overlay circularly, 7 colors with white background from start to end",
    70: "Overlay circularly, 7 colors with black background from end to start",
    71: "Overlay circularly, 7 colors with red background from end to start",
    72: "Overlay circularly, 7 colors with green background from end to start",
    73: "Overlay circularly, 7 colors with blue background from end to start",
    74: "Overlay circularly, 7 colors with yellow background from end to start",
    75: "Overlay circularly, 7 colors with purple background from end to start",
    76: "Overlay circularly, 7 colors with cyan background from end to start",
    77: "Overlay circularly, 7 colors with white background from end to start",
    78: "Overlay circularly, 7 colors with black background from start to end and return back",
    79: "Overlay circularly, 7 colors with red background from start to end and return back",
    80: "Overlay circularly, 7 colors with green background from start to end and return back",
    81: "Overlay circularly, 7 colors with blue background from start to end and return back",
    82: "Overlay circularly, 7 colors with yellow background from start to end and return back",
    83: "Overlay circularly, 7 colors with purple background from start to end and return back",
    84: "Overlay circularly, 7 colors with cyan background from start to end and return back",
    85: "Overlay circularly, 7 colors with white background from start to end and return back",
    86: "Overlay circularly, 7 colors with black background from middle to both ends",
    87: "Overlay circularly, 7 colors with red background from middle to both ends",
    88: "Overlay circularly, 7 colors with green background from middle to both ends",
    89: "Overlay circularly, 7 colors with blue background from middle to both ends",
    90: "Overlay circularly, 7 colors with yellow background from middle to both ends",
    91: "Overlay circularly, 7 colors with purple background from middle to both ends",
    92: "Overlay circularly, 7 colors with cyan background from middle to both ends",
    93: "Overlay circularly, 7 colors with white background from middle to both ends",
    94: "Overlay circularly, 7 colors with black background from both ends to middle",
    95: "Overlay circularly, 7 colors with red background from both ends to middle",
    96: "Overlay circularly, 7 colors with green background from both ends to middle",
    97: "Overlay circularly, 7 colors with blue background from both ends to middle",
    98: "Overlay circularly, 7 colors with yellow background from both ends to middle",
    99: "Overlay circularly, 7 colors with purple background from both ends to middle",
    100: "Overlay circularly, 7 colors with cyan background from both ends to middle",
    101: "Overlay circularly, 7 colors with white background from both ends to middle",
    102: "Overlay circularly, 7 colors with black background from middle to both sides and return back",
    103: "Overlay circularly, 7 colors with red background from middle to both sides and return back",
    104: "Overlay circularly, 7 colors with green background from middle to both sides and return back",
    105: "Overlay circularly, 7 colors with blue background from middle to both sides and return back",
    106: "Overlay circularly, 7 colors with yellow background from middle to both sides and return back",
    107: "Overlay circularly, 7 colors with purple background from middle to both sides and return back",
    108: "Overlay circularly, 7 colors with cyan background from middle to both sides and return back",
    109: "Overlay circularly, 7 colors with white background from middle to both sides and return back",
    110: "Fading out run circularly, 1point with black background from start to end",
    111: "Fading out run circularly, 1point with red background from start to end",
    112: "Fading out run circularly, 1point with green background from start to end",
    113: "Fading out run circularly, 1point with blue background from start to end",
    114: "Fading out run circularly, 1point with yellow background from start to end",
    115: "Fading out run circularly, 1point with purple background from start to end",
    116: "Fading out run circularly, 1point with cyan background from start to end",
    117: "Fading out run circularly, 1point with white background from start to end",
    118: "Fading out run circularly, 1point with black background from end to start",
    119: "Fading out run circularly, 1point with red background from end to start",
    120: "Fading out run circularly, 1point with green background from end to start",
    121: "Fading out run circularly, 1point with blue background from end to start",
    122: "Fading out run circularly, 1point with yellow background from end to start",
    123: "Fading out run circularly, 1point with purple background from end to start",
    124: "Fading out run circularly, 1point with cyan background from end to start",
    125: "Fading out run circularly, 1point with white background from end to start",
    126: "Fading out run circularly, 1point with black background from start to end and return back",
    127: "Fading out run circularly, 1point with red background from start to end and return back",
    128: "Fading out run circularly, 1point with green background from start to end and return back",
    129: "Fading out run circularly, 1point with blue background from start to end and return back",
    130: "Fading out run circularly, 1point with yellow background from start to end and return back",
    131: "Fading out run circularly, 1point with purple background from start to end and return back",
    132: "Fading out run circularly, 1point with cyan background from start to end and return back",
    133: "Fading out run circularly, 1point with white background from start to end and return back",
    134: "Flows in olivary circularly, 7 colors with black background from start to end",
    135: "Flows in olivary circularly, 7 colors with red background from start to end",
    136: "Flows in olivary circularly, 7 colors with green background from start to end",
    137: "Flows in olivary circularly, 7 colors with blue background from start to end",
    138: "Flows in olivary circularly, 7 colors with yellow background from start to end",
    139: "Flows in olivary circularly, 7 colors with purple background from start to end",
    140: "Flows in olivary circularly, 7 colors with cyan background from start to end",
    141: "Flows in olivary circularly, 7 colors with white background from start to end",
    142: "Flows in olivary circularly, 7 colors with black background from end to start",
    143: "Flows in olivary circularly, 7 colors with red background from end to start",
    144: "Flows in olivary circularly, 7 colors with green background from end to start",
    145: "Flows in olivary circularly, 7 colors with blue background from end to start",
    146: "Flows in olivary circularly, 7 colors with yellow background from end to start",
    147: "Flows in olivary circularly, 7 colors with purple background from end to start",
    148: "Flows in olivary circularly, 7 colors with cyan background from end to start",
    149: "Flows in olivary circularly, 7 colors with white background from end to start",
    150: "Flows in olivary circularly, 7 colors with black background from start to end and return back",
    151: "Flows in olivary circularly, 7 colors with red background from start to end and return back",
    152: "Flows in olivary circularly, 7 colors with green background from start to end and return back",
    153: "Flows in olivary circularly, 7 colors with blue background from start to end and return back",
    154: "Flows in olivary circularly, 7 colors with yellow background from start to end and return back",
    155: "Flows in olivary circularly, 7 colors with purple background from start to end and return back",
    156: "Flows in olivary circularly, 7 colors with cyan background from start to end and return back",
    157: "Flows in olivary circularly, 7 colors with white background from start to end and return back",
    158: "7 colors run circularly, each color in every 1 point with black background from start to end",
    159: "7 colors run circularly, each color in every 1 point with red background from start to end",
    160: "7 colors run circularly, each color in every 1 point with green background from start to end",
    161: "7 colors run circularly, each color in every 1 point with blue background from start to end",
    162: "7 colors run circularly, each color in every 1 point with yellow background from start to end",
    163: "7 colors run circularly, each color in every 1 point with purple background from start to end",
    164: "7 colors run circularly, each color in every 1 point with cyan background from start to end",
    165: "7 colors run circularly, each color in every 1 point with white background from start to end",
    166: "7 colors run circularly, each color in every 1 point with black background from end to start",
    167: "7 colors run circularly, each color in every 1 point with red background from end to start",
    168: "7 colors run circularly, each color in every 1 point with green background from end to start",
    169: "7 colors run circularly, each color in every 1 point with blue background from end to start",
    170: "7 colors run circularly, each color in every 1 point with yellow background from end to start",
    171: "7 colors run circularly, each color in every 1 point with purple background from end to start",
    172: "7 colors run circularly, each color in every 1 point with cyan background from end to start",
    173: "7 colors run circularly, each color in every 1 point with white background from end to start",
    174: "7 colors run circularly, each color in every 1 point with black background from start to end and return back",
    175: "7 colors run circularly, each color in every 1 point with red background from start to end and return back",
    176: "7 colors run circularly, each color in every 1 point with green background from start to end and return back",
    177: "7 colors run circularly, each color in every 1 point with blue background from start to end and return back",
    178: "7 colors run circularly, each color in every 1 point with yellow background from start to end and return back",
    179: "7 colors run circularly, each color in every 1 point with purple background from start to end and return back",
    180: "7 colors run circularly, each color in every 1 point with cyan background from start to end and return back",
    181: "7 colors run circularly, each color in every 1 point with white background from start to end and return back",
    182: "7 colors run circularly, each color in multi points with red background from start to end",
    183: "7 colors run circularly, each color in multi points with green background from start to end",
    184: "7 colors run circularly, each color in multi points with blue background from start to end",
    185: "7 colors run circularly, each color in multi points with yellow background from start to end",
    186: "7 colors run circularly, each color in multi points with purple background from start to end",
    187: "7 colors run circularly, each color in multi points with cyan background from start to end",
    188: "7 colors run circularly, each color in multi points with white background from start to end",
    189: "7 colors run circularly, each color in multi points with red background from end to start",
    190: "7 colors run circularly, each color in multi points with green background from end to start",
    191: "7 colors run circularly, each color in multi points with blue background from end to start",
    192: "7 colors run circularly, each color in multi points with yellow background from end to start",
    193: "7 colors run circularly, each color in multi points with purple background from end to start",
    194: "7 colors run circularly, each color in multi points with cyan background from end to start",
    195: "7 colors run circularly, each color in multi points with white background from end to start",
    196: "7 colors run circularly, each color in multi points with red background from start to end and return back",
    197: "7 colors run circularly, each color in multi points with green background from start to end and return back",
    198: "7 colors run circularly, each color in multi points with blue background from start to end and return back",
    199: "7 colors run circularly, each color in multi points with yellow background from start to end and return back",
    200: "7 colors run circularly, each color in multi points with purple background from start to end and return back",
    201: "7 colors run circularly, each color in multi points with cyan background from start to end and return back",
    202: "7 colors run circularly, each color in multi points with white background from start to end and return back",
    203: "Fading out run circularly, 7 colors each in red fading from start to end",
    204: "Fading out run circularly, 7 colors each in green fading from start to end",
    205: "Fading out run circularly, 7 colors each in blue fading from start to end",
    206: "Fading out run circularly, 7 colors each in yellow fading from start to end",
    207: "Fading out run circularly, 7 colors each in purple fading from start to end",
    208: "Fading out run circularly, 7 colors each in cyan fading from start to end",
    209: "Fading out run circularly, 7 colors each in white fading from start to end",
    210: "Fading out run circularly, 7 colors each in red fading from end to start",
    211: "Fading out run circularly, 7 colors each in green fading from end to start",
    212: "Fading out run circularly, 7 colors each in blue fading from end to start",
    213: "Fading out run circularly, 7 colors each in yellow fading from end to start",
    214: "Fading out run circularly, 7 colors each in purple fading from end to start",
    215: "Fading out run circularly, 7 colors each in cyan fading from end to start",
    216: "Fading out run circularly, 7 colors each in white fading from end to start",
    217: "Fading out run circularly, 7 colors each in red fading from start to end and return back",
    218: "Fading out run circularly, 7 colors each in green fading from start to end and return back",
    219: "Fading out run circularly, 7 colors each in blue fading from start to end and return back",
    220: "Fading out run circularly, 7 colors each in yellow fading from start to end and return back",
    221: "Fading out run circularly, 7 colors each in purple fading from start to end and return back",
    222: "Fading out run circularly, 7 colors each in cyan fading from start to end and return back",
    223: "Fading out run circularly, 7 colors each in white fading from start to end and return back",
    224: "7 colors each in red run circularly, multi points from start to end",
    225: "7 colors each in green run circularly, multi points from start to end",
    226: "7 colors each in blue run circularly, multi points from start to end",
    227: "7 colors each in yellow run circularly, multi points from start to end",
    228: "7 colors each in purple run circularly, multi points from start to end",
    229: "7 colors each in cyan run circularly, multi points from start to end",
    230: "7 colors each in white run circularly, multi points from start to end",
    231: "7 colors each in red run circularly, multi points from end to start",
    232: "7 colors each in green run circularly, multi points from end to start",
    233: "7 colors each in blue run circularly, multi points from end to start",
    234: "7 colors each in yellow run circularly, multi points from end to start",
    235: "7 colors each in purple run circularly, multi points from end to start",
    236: "7 colors each in cyan run circularly, multi points from end to start",
    237: "7 colors each in white run circularly, multi points from end to start",
    238: "7 colors each in red run circularly, multi points from start to end and return back",
    239: "7 colors each in green run circularly, multi points from start to end and return back",
    240: "and return back7 colors each in blue run circularly, multi points from start to end",
    241: "7 colors each in yellow run circularly, multi points from start to end and return back",
    242: "7 colors each in purple run circularly, multi points from start to end and return back",
    243: "7 colors each in cyan run circularly, multi points from start to end and return back",
    244: "7 colors each in white run circularly, multi points from start to end and return back",
    245: "Flows gradually and circularly, 6 colors with red background from start to end",
    246: "Flows gradually and circularly, 6 colors with green background from start to end",
    247: "Flows gradually and circularly, 6 colors with blue background from start to end",
    248: "Flows gradually and circularly, 6 colors with yellow background from start to end",
    249: "Flows gradually and circularly, 6 colors with purple background from start to end",
    250: "Flows gradually and circularly, 6 colors with cyan background from start to end",
    251: "Flows gradually and circularly, 6 colors with white background from start to end",
    252: "Flows gradually and circularly, 6 colors with red background from end to start",
    253: "Flows gradually and circularly, 6 colors with green background from end to start",
    254: "Flows gradually and circularly, 6 colors with blue background from end to start",
    255: "Flows gradually and circularly, 6 colors with yellow background from end to start",
    256: "Flows gradually and circularly, 6 colors with purple background from end to start",
    257: "Flows gradually and circularly, 6 colors with cyan background from end to start",
    258: "Flows gradually and circularly, 6 colors with white background from end to start",
    259: "Flows gradually and circularly, 6 colors with red background from start to end and return back",
    260: "Flows gradually and circularly, 6 colors with green background from start to end and return back",
    261: "Flows gradually and circularly, 6 colors with blue background from start to end and return back",
    262: "Flows gradually and circularly, 6 colors with yellow background from start to end and return back",
    263: "Flows gradually and circularly, 6 colors with purple background from start to end and return back",
    264: "Flows gradually and circularly, 6 colors with cyan background from start to end and return back",
    265: "Flows gradually and circularly, 6 colors with white background from start to end and return back",
    266: "7 colors run with black background from start to end",
    267: "7 colors run with red background from start to end",
    268: "7 colors run with green background from start to end",
    269: "7 colors run with blue background from start to end",
    270: "7 colors run with yellow background from start to end",
    271: "7 colors run with purple background from start to end",
    272: "7 colors run with cyan background from start to end",
    273: "7 colors run with white background from start to end",
    274: "7 colors run with black background from end to start",
    275: "7 colors run with red background from end to start",
    276: "7 colors run with green background from end to start",
    277: "7 colors run with blue background from end to start",
    278: "7 colors run with yellow background from end to start",
    279: "7 colors run with purple background from end to start",
    280: "7 colors run with cyan background from end to start",
    281: "7 colors run with white background from end to start",
    282: "7 colors run with black background from start to end and return back",
    283: "7 colors run with red background from start to end and return back",
    284: "7 colors run with green background from start to end and return back",
    285: "7 colors run with blue background from start to end and return back",
    286: "7 colors run with yellow background from start to end and return back",
    287: "7 colors run with purple background from start to end and return back",
    288: "7 colors run with cyan background from start to end and return back",
    289: "7 colors run with white background from start to end and return back",
    290: "7 colors run gradually + 7 colors run in olivary",
    291: "7 colors run gradually + 7 colors change quickly",
    292: "7 colors run gradually + 7 colors flash",
    293: "7 colors run in olivary + 7 colors change quickly",
    294: "7 colors run in olivary + 7 colors flash",
    295: "7 colors change quickly + 7 colors flash",
    296: "7 colors run gradually + 7 colors run in olivary + 7 colors change quickly",
    297: "7 colors run gradually + 7 colors run in olivary + 7 colors flash",
    298: "7 colors run gradually + 7 colors change quickly + 7 colors flash",
    299: "7 colors run in olivary + 7 colors change quickly + 7 colors flash",
    300: "7 colors run gradually + 7 colors run in olivary + 7 colors change quickly + 7 color flash",
}
ORIGINAL_ADDRESSABLE_EFFECT_NAME_ID = {
    v: k for k, v in ORIGINAL_ADDRESSABLE_EFFECT_ID_NAME.items()
}


class PresetPattern:

    _instance = None
    seven_color_cross_fade = 0x25
    red_gradual_change = 0x26
    green_gradual_change = 0x27
    blue_gradual_change = 0x28
    yellow_gradual_change = 0x29
    cyan_gradual_change = 0x2A
    purple_gradual_change = 0x2B
    white_gradual_change = 0x2C
    red_green_cross_fade = 0x2D
    red_blue_cross_fade = 0x2E
    green_blue_cross_fade = 0x2F
    seven_color_strobe_flash = 0x30
    red_strobe_flash = 0x31
    green_strobe_flash = 0x32
    blue_strobe_flash = 0x33
    yellow_strobe_flash = 0x34
    cyan_strobe_flash = 0x35
    purple_strobe_flash = 0x36
    white_strobe_flash = 0x37
    seven_color_jumping = 0x38

    def __init__(self) -> None:
        self._value_to_str: Dict[int, str] = {
            v: k.replace("_", " ").title()
            for k, v in PresetPattern.__dict__.items()
            if type(v) is int
        }

    @classmethod
    def instance(cls) -> "PresetPattern":
        """Get preset pattern instance."""
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @staticmethod
    def valid(pattern: int) -> bool:
        if pattern >= 0x25 and pattern <= 0x38 or pattern >= 0x61 and pattern <= 0x63:
            return True
        return False

    @staticmethod
    def valtostr(pattern: int) -> Optional[str]:
        instance = PresetPattern.instance()
        return instance._value_to_str.get(pattern)

    @staticmethod
    def str_to_val(effect: str) -> int:
        if effect in EFFECT_MAP:
            return EFFECT_MAP[effect]
        if hasattr(PresetPattern, effect):
            return cast(int, getattr(PresetPattern, effect))
        raise ValueError(f"{effect} is not a known effect name.")



import asyncio
import contextlib
from datetime import date
import logging
import select
import socket
import sys
import time
from typing import Any, Dict, List, Optional, Tuple, Union, cast

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict

from .const import (
    ATTR_FIRMWARE_DATE,
    ATTR_ID,
    ATTR_IPADDR,
    ATTR_MODEL,
    ATTR_MODEL_DESCRIPTION,
    ATTR_MODEL_INFO,
    ATTR_MODEL_NUM,
    ATTR_VERSION_NUM,
)
from .models_db import get_model_description

_LOGGER = logging.getLogger(__name__)


class FluxLEDDiscovery(TypedDict):
    """A flux led device."""

    ipaddr: str
    id: Optional[str]  # aka mac
    model: Optional[str]
    model_num: Optional[int]
    version_num: Optional[int]
    firmware_date: Optional[date]
    model_info: Optional[str]
    model_description: Optional[str]


def _process_discovery_message(data: Dict[str, Any], decoded_data: str) -> None:
    """Process response from b'HF-A11ASSISTHREAD'

    b'192.168.214.252,B4E842E10588,AK001-ZJ2145'
    """
    data_split = decoded_data.split(",")
    if len(data_split) < 3:
        return
    ipaddr = data_split[0]
    data.update(
        {
            ATTR_IPADDR: ipaddr,
            ATTR_ID: data_split[1],
            ATTR_MODEL: data_split[2],
        }
    )


def _process_version_message(data: Dict[str, Any], decoded_data: str) -> None:
    """Process response from b'AT+LVER\r'

    b'+ok=07_06_20210106_ZG-BL\r'
    """
    version_data = decoded_data[4:].replace("\r", "")
    data_split = version_data.split("_")
    if len(data_split) < 2:
        return
    try:
        data[ATTR_MODEL_NUM] = int(data_split[0], 16)
        data[ATTR_VERSION_NUM] = int(data_split[1], 16)
    except ValueError:
        return
    data[ATTR_MODEL_DESCRIPTION] = get_model_description(data[ATTR_MODEL_NUM])
    if len(data_split) < 3:
        return
    firmware_date = data_split[2]
    try:
        data[ATTR_FIRMWARE_DATE] = date(
            int(firmware_date[:4]),
            int(firmware_date[4:6]),
            int(firmware_date[6:8]),
        )
    except (TypeError, ValueError):
        return
    if len(data_split) < 4:
        return
    data[ATTR_MODEL_INFO] = data_split[3]


class BulbScanner:

    DISCOVERY_PORT = 48899
    BROADCAST_FREQUENCY = (
        5  # At least 5 for A1 models (Magic Home Branded RGB Symphony [Addressable])
    )
    RESPONSE_SIZE = 64
    DISCOVER_MESSAGE = b"HF-A11ASSISTHREAD"
    VERSION_MESSAGE = b"AT+LVER\r"
    BROADCAST_ADDRESS = "<broadcast>"

    def __init__(self) -> None:
        self.found_bulbs: List[FluxLEDDiscovery] = []

    def getBulbInfoByID(self, id: str) -> FluxLEDDiscovery:
        for b in self.found_bulbs:
            if b["id"] == id:
                return b
        return b

    def getBulbInfo(self) -> List[FluxLEDDiscovery]:
        return self.found_bulbs

    def _create_socket(self) -> socket.socket:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        with contextlib.suppress(Exception):
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        sock.bind(("", self.DISCOVERY_PORT))
        sock.setblocking(False)
        return sock

    def _destination_from_address(self, address: Optional[str]) -> Tuple[str, int]:
        if address is None:
            address = self.BROADCAST_ADDRESS
        return (address, self.DISCOVERY_PORT)

    def _process_response(
        self,
        data: Optional[bytes],
        from_address: Tuple[str, int],
        address: Optional[str],
        response_list: Dict[str, Dict[str, Any]],
    ) -> bool:
        """Process a response.

        Returns True if processing should stop
        """
        if data is None:
            return False
        if data in (self.DISCOVER_MESSAGE, self.VERSION_MESSAGE):
            return False
        decoded_data = data.decode("ascii")
        self._process_data(from_address, decoded_data, response_list)
        if address is None:
            return False
        return response_list.get(address, {}).get(ATTR_MODEL_NUM) is not None

    def _process_data(
        self,
        from_address: Tuple[str, int],
        decoded_data: str,
        response_list: Dict[str, Dict[str, Any]],
    ) -> None:
        """Process data."""
        from_ipaddr = from_address[0]
        data = response_list.setdefault(from_ipaddr, {})
        if decoded_data.startswith("+ok="):
            _process_version_message(data, decoded_data)
        elif "," in decoded_data:
            _process_discovery_message(data, decoded_data)

    def _found_bulbs(
        self, response_list: Dict[str, Dict[str, Any]]
    ) -> List[FluxLEDDiscovery]:
        """Return only complete bulb discoveries."""

        return [
            cast(FluxLEDDiscovery, info)
            for info in response_list.values()
            if info.get(ATTR_IPADDR)
        ]

    def send_discovery_messages(
        self,
        sender: Union[socket.socket, asyncio.DatagramTransport],
        destination: Tuple[str, int],
    ) -> None:
        _LOGGER.debug("discover: %s => %s", destination, self.DISCOVER_MESSAGE)
        sender.sendto(self.DISCOVER_MESSAGE, destination)
        _LOGGER.debug("discover: %s => %s", destination, self.VERSION_MESSAGE)
        sender.sendto(self.VERSION_MESSAGE, destination)

    def scan(
        self, timeout: int = 10, address: Optional[str] = None
    ) -> List[FluxLEDDiscovery]:
        """Scan for bulbs.

        If an address is provided, the scan will return
        as soon as it gets a response from that address
        """
        sock = self._create_socket()
        destination = self._destination_from_address(address)
        # set the time at which we will quit the search
        quit_time = time.monotonic() + timeout
        response_list: Dict[str, Dict[str, Any]] = {}
        found_all = False
        # outer loop for query send
        while not found_all:
            if time.monotonic() > quit_time:
                break
            # send out a broadcast query
            self.send_discovery_messages(sock, destination)
            # inner loop waiting for responses
            while True:
                sock.settimeout(1)
                remain_time = quit_time - time.monotonic()
                time_out = min(remain_time, timeout / self.BROADCAST_FREQUENCY)
                if time_out <= 0:
                    break
                read_ready, _, _ = select.select([sock], [], [], time_out)
                if not read_ready:
                    if time.monotonic() < quit_time:
                        # No response, send broadcast again in cast it got lost
                        self.send_discovery_messages(sock, destination)
                    continue

                try:
                    data, addr = sock.recvfrom(self.RESPONSE_SIZE)
                    _LOGGER.debug("discover: %s <= %s", addr, data)
                except socket.timeout:
                    continue

                if self._process_response(data, addr, address, response_list):
                    found_all = True
                    break

        self.found_bulbs = self._found_bulbs(response_list)
        return self.found_bulbs



#!/usr/bin/env python
"""
This is a utility for controlling stand-alone Flux WiFi LED light bulbs.

The protocol was reverse-engineered by studying packet captures between a
bulb and the controlling "Magic Home" mobile app.  The code here dealing

with the network protocol is littered with magic numbers, and ain't so pretty.
But it does seem to work!

So far most of the functionality of the apps is available here via the CLI
and/or programmatically.

The classes in this project could very easily be used as an API, and incorporated into a GUI app written
in PyQt, Kivy, or some other framework.

##### Available:
* Discovering bulbs on LAN
* Turning on/off bulb
* Get state information
* Setting "warm white" mode
* Setting single color mode
* Setting preset pattern mode
* Setting custom pattern mode
* Reading timers
* Setting timers

##### Some missing pieces:
* Initial administration to set up WiFi SSID and passphrase/key.
* Remote access administration
* Music-relating pulsing. This feature isn't so impressive on the Magic Home app,
and looks like it might be a bit of work.

##### Cool feature:
* Specify colors with names or web hex values.  Requires that python "webcolors"
package is installed.  (Easily done via pip, easy_install, or apt-get, etc.)
See the following for valid color names: http://www.w3schools.com/html/html_colornames.asp

"""

import datetime
import logging
from optparse import OptionGroup, OptionParser, Values
import sys
from typing import Any, List, Optional, Tuple

from .device import WifiLedBulb
from .pattern import PresetPattern
from .scanner import BulbScanner, FluxLEDDiscovery
from .timer import LedTimer
from .utils import utils

_LOGGER = logging.getLogger(__name__)


# =======================================================================
def showUsageExamples() -> None:
    example_text = """
Examples:

Scan network:
    %prog% -s

Scan network and show info about all:
    %prog% -sSti

Turn on:
    %prog% 192.168.1.100 --on
    %prog% 192.168.1.100 -192.168.1.101 -1

Turn on all bulbs on LAN:
    %prog% -sS --on

Turn off:
    %prog% 192.168.1.100 --off
    %prog% 192.168.1.100 --0
    %prog% -sS --off

Set warm white, 75%
    %prog% 192.168.1.100 -w 75

Set cold white, 55%
    %prog% 192.168.1.100 -d 55

Set CCT, 3500 85%
    %prog% 192.168.1.100 -k 3500 85

Set fixed color red :
    %prog% 192.168.1.100 -c Red
    %prog% 192.168.1.100 -c 255,0,0
    %prog% 192.168.1.100 -c "#FF0000"

Set RGBW 25 100 200 50:
    %prog% 192.168.1.100 -c 25,100,200,50

Set RGBWW 25 100 200 50 30:
    %prog% 192.168.1.100 -c 25,100,200,50,30

Set preset pattern #35 with 40% speed:
    %prog% 192.168.1.100 -p 35 40

Set custom pattern 25% speed, red/green/blue, gradual change:
    %prog% 192.168.1.100 -C gradual 25 "red green (0,0,255)"

Sync all bulb's clocks with this computer's:
    %prog% -sS --setclock

Set timer #1 to turn on red at 5:30pm on weekdays:
    %prog% 192.168.1.100 -T 1 color "time:1730;repeat:12345;color:red"

Deactivate timer #4:
    %prog% 192.168.1.100 -T 4 inactive ""

Use --timerhelp for more details on setting timers
    """

    print(example_text.replace("%prog%", sys.argv[0]))


def showTimerHelp() -> None:
    timerhelp_text = """
    There are 6 timers available for each bulb.

Mode Details:
    inactive:   timer is inactive and unused
    poweroff:   turns off the light
    default:    turns on the light in default mode
    color:      turns on the light with specified color
    preset:     turns on the light with specified preset and speed
    warmwhite:  turns on the light with warm white at specified brightness

Settings available for each mode:
    Timer Mode | Settings
    --------------------------------------------
    inactive:   [none]
    poweroff:   time, (repeat | date)
    default:    time, (repeat | date)
    color:      time, (repeat | date), color
    preset:     time, (repeat | date), code, speed
    warmwhite:  time, (repeat | date), level
    sunrise:    time, (repeat | date), startBrightness, endBrightness, duration
    sunset:     time, (repeat | date), startBrightness, endBrightness, duration

Setting Details:

    time: 4 digit string with zeros, no colons
        e.g:
        "1000"  - for 10:00am
        "2312"  - for 11:23pm
        "0315"  - for 3:15am

    repeat: Days of the week that the timer should repeat
            (Mutually exclusive with date)
            0=Sun, 1=Mon, 2=Tue, 3=Wed, 4=Thu, 5=Fri, 6=Sat
        e.g:
        "0123456"  - everyday
        "06"       - weekends
        "12345"    - weekdays
        "2"        - only Tuesday

    date: Date that the one-time timer should fire
            (Mutually exclusive with repeat)
        e.g:
        "2015-09-13"
        "2016-12-03"

    color: Color name, hex code, or rgb triple

    level: Level of the warm while light (0-100)

    code:  Code of the preset pattern (use -l to list them)

    speed: Speed of the preset pattern transitions (0-100)

    startBrightness: starting brightness of warmlight (0-100)

    endBrightness: ending brightness of warmlight (0-100)

    duration: transition time in minutes

Example setting strings:
    "time:2130;repeat:0123456"
    "time:2130;date:2015-08-11"
    "time:1245;repeat:12345;color:123,345,23"
    "time:1245;repeat:12345;color:green"
    "time:1245;repeat:06;code:50;speed:30"
    "time:0345;date:2015-08-11;level:100"
    """

    print(timerhelp_text)


def processSetTimerArgs(parser: OptionParser, args: Any) -> LedTimer:  # noqa: C901
    mode = args[1]
    num = args[0]
    settings = args[2]

    if not num.isdigit() or int(num) > 6 or int(num) < 1:
        parser.error("Timer number must be between 1 and 6")

    # create a dict from the settings string
    settings_list = settings.split(";")
    settings_dict = {}
    for s in settings_list:
        pair = s.split(":")
        key = pair[0].strip().lower()
        val = ""
        if len(pair) > 1:
            val = pair[1].strip().lower()
        settings_dict[key] = val

    keys = list(settings_dict.keys())
    timer = LedTimer()

    if mode == "inactive":
        # no setting needed
        timer.setActive(False)

    elif mode in [
        "poweroff",
        "default",
        "color",
        "preset",
        "warmwhite",
        "sunrise",
        "sunset",
    ]:
        timer.setActive(True)

        if "time" not in keys:
            parser.error(f"This mode needs a time: {mode}")
        if "repeat" in keys and "date" in keys:
            parser.error(f"This mode only a repeat or a date, not both: {mode}")

        # validate time format
        if len(settings_dict["time"]) != 4 or not settings_dict["time"].isdigit():
            parser.error("time must be a 4 digits")
        hour = int(settings_dict["time"][0:2:])
        minute = int(settings_dict["time"][2:4:])
        if hour > 23:
            parser.error("timer hour can't be greater than 23")
        if minute > 59:
            parser.error("timer minute can't be greater than 59")

        timer.setTime(hour, minute)

        # validate date format
        if "repeat" not in keys and "date" not in keys:
            # Generate date for next occurance of time
            print("No time or repeat given. Defaulting to next occurance of time")
            now = datetime.datetime.now()
            dt = now.replace(hour=hour, minute=minute)
            if utils.date_has_passed(dt):
                dt = dt + datetime.timedelta(days=1)
            # settings_dict["date"] = date
            timer.setDate(dt.year, dt.month, dt.day)
        elif "date" in keys:
            try:
                dt = datetime.datetime.strptime(settings_dict["date"], "%Y-%m-%d")
                timer.setDate(dt.year, dt.month, dt.day)
            except ValueError:
                parser.error("date is not properly formatted: YYYY-MM-DD")

        # validate repeat format
        if "repeat" in keys:
            if len(settings_dict["repeat"]) == 0:
                parser.error("Must specify days to repeat")
            days = set()
            for c in list(settings_dict["repeat"]):
                if c not in ["0", "1", "2", "3", "4", "5", "6"]:
                    parser.error("repeat can only contain digits 0-6")
                days.add(int(c))

            repeat = 0
            if 0 in days:
                repeat |= LedTimer.Su
            if 1 in days:
                repeat |= LedTimer.Mo
            if 2 in days:
                repeat |= LedTimer.Tu
            if 3 in days:
                repeat |= LedTimer.We
            if 4 in days:
                repeat |= LedTimer.Th
            if 5 in days:
                repeat |= LedTimer.Fr
            if 6 in days:
                repeat |= LedTimer.Sa
            timer.setRepeatMask(repeat)

        if mode == "default":
            timer.setModeDefault()

        if mode == "poweroff":
            timer.setModeTurnOff()

        if mode == "color":
            if "color" not in keys:
                parser.error("color mode needs a color setting")
            # validate color val
            c = utils.color_object_to_tuple(settings_dict["color"])  # type: ignore
            if c is None:
                parser.error("Invalid color value: {}".format(settings_dict["color"]))
            assert c is not None
            timer.setModeColor(c[0], c[1], c[2])  # type: ignore

        if mode == "preset":
            if "code" not in keys:
                parser.error(f"preset mode needs a code: {mode}")
            if "speed" not in keys:
                parser.error(f"preset mode needs a speed: {mode}")
            code = settings_dict["code"]
            speed = settings_dict["speed"]
            if not speed.isdigit() or int(speed) > 100:
                parser.error("preset speed must be a percentage (0-100)")
            if not code.isdigit() or not PresetPattern.valid(int(code)):
                parser.error("preset code must be in valid range")
            timer.setModePresetPattern(int(code), int(speed))

        if mode == "warmwhite":
            if "level" not in keys:
                parser.error(f"warmwhite mode needs a level: {mode}")
            level = settings_dict["level"]
            if not level.isdigit() or int(level) > 100:
                parser.error("warmwhite level must be a percentage (0-100)")
            timer.setModeWarmWhite(int(level))

        if mode == "sunrise" or mode == "sunset":
            if "startbrightness" not in keys:
                parser.error(f"{mode} mode needs a startBrightness (0% -> 100%)")
            startBrightness = int(settings_dict["startbrightness"])

            if "endbrightness" not in keys:
                parser.error(f"{mode} mode needs an endBrightness (0% -> 100%)")
            endBrightness = int(settings_dict["endbrightness"])

            if "duration" not in keys:
                parser.error(f"{mode} mode needs a duration (minutes)")
            duration = int(settings_dict["duration"])

            if mode == "sunrise":
                timer.setModeSunrise(startBrightness, endBrightness, duration)

            elif mode == "sunset":
                timer.setModeSunset(startBrightness, endBrightness, duration)

    else:
        parser.error(f"Not a valid timer mode: {mode}")

    return timer


def processCustomArgs(
    parser: OptionParser, args: Any
) -> Optional[Tuple[Any, int, List[Tuple[int, ...]]]]:
    if args[0] not in ["gradual", "jump", "strobe"]:
        parser.error(f"bad pattern type: {args[0]}")
        return None

    speed = int(args[1])

    # convert the string to a list of RGB tuples
    # it should have space separated items of either
    # color names, hex values, or byte triples
    try:
        color_list_str = args[2].strip()
        str_list = color_list_str.split(" ")
        color_list = []
        for s in str_list:
            c = utils.color_object_to_tuple(s)
            if c is not None:
                color_list.append(c)
            else:
                raise Exception

    except Exception:
        parser.error(
            "COLORLIST isn't formatted right.  It should be a space separated list of RGB tuples, color names or web hex values"
        )

    return args[0], speed, color_list


def parseArgs() -> Tuple[Values, Any]:  # noqa: C901

    parser = OptionParser()

    parser.description = "A utility to control Flux WiFi LED Bulbs. "
    # parser.description += ""
    # parser.description += "."
    power_group = OptionGroup(parser, "Power options (mutually exclusive)")
    mode_group = OptionGroup(parser, "Mode options (mutually exclusive)")
    info_group = OptionGroup(parser, "Program help and information option")
    other_group = OptionGroup(parser, "Other options")

    parser.add_option_group(info_group)
    info_group.add_option(
        "-e",
        "--examples",
        action="store_true",
        dest="showexamples",
        default=False,
        help="Show usage examples",
    )
    info_group.add_option(
        "",
        "--timerhelp",
        action="store_true",
        dest="timerhelp",
        default=False,
        help="Show detailed help for setting timers",
    )
    info_group.add_option(
        "-l",
        "--listpresets",
        action="store_true",
        dest="listpresets",
        default=False,
        help="List preset codes",
    )
    info_group.add_option(
        "--listcolors",
        action="store_true",
        dest="listcolors",
        default=False,
        help="List color names",
    )

    parser.add_option(
        "-s",
        "--scan",
        action="store_true",
        dest="scan",
        default=False,
        help="Search for bulbs on local network",
    )
    parser.add_option(
        "-S",
        "--scanresults",
        action="store_true",
        dest="scanresults",
        default=False,
        help="Operate on scan results instead of arg list",
    )
    power_group.add_option(
        "-1",
        "--on",
        action="store_true",
        dest="on",
        default=False,
        help="Turn on specified bulb(s)",
    )
    power_group.add_option(
        "-0",
        "--off",
        action="store_true",
        dest="off",
        default=False,
        help="Turn off specified bulb(s)",
    )
    parser.add_option_group(power_group)

    mode_group.add_option(
        "-c",
        "--color",
        dest="color",
        default=None,
        help="""For setting a single color mode.  Can be either color name, web hex, or comma-separated RGB triple.
        For setting an RGBW can be a comma-seperated RGBW list
        For setting an RGBWW can be a comma-seperated RGBWW list""",
        metavar="COLOR",
    )
    mode_group.add_option(
        "-w",
        "--warmwhite",
        dest="ww",
        default=None,
        help="Set warm white mode (LEVELWW is percent)",
        metavar="LEVELWW",
        type="int",
    )
    mode_group.add_option(
        "-d",
        "--coldwhite",
        dest="cw",
        default=None,
        help="Set cold white mode (LEVELCW is percent)",
        metavar="LEVELCW",
        type="int",
    )
    mode_group.add_option(
        "-k",
        "--CCT",
        dest="cct",
        default=None,
        help="Temperture and brightness (CCT Kelvin, brightness percent)",
        metavar="LEVELCCT",
        type="int",
        nargs=2,
    )
    mode_group.add_option(
        "-p",
        "--preset",
        dest="preset",
        default=None,
        help="Set preset pattern mode (SPEED is percent)",
        metavar="CODE SPEED",
        type="int",
        nargs=2,
    )
    mode_group.add_option(
        "-C",
        "--custom",
        dest="custom",
        metavar="TYPE SPEED COLORLIST",
        default=None,
        nargs=3,
        help="Set custom pattern mode. "
        + "TYPE should be jump, gradual, or strobe. SPEED is percent. "
        + "COLORLIST is a space-separated list of color names, web hex values, or comma-separated RGB triples",
    )
    parser.add_option_group(mode_group)

    parser.add_option(
        "-i",
        "--info",
        action="store_true",
        dest="info",
        default=False,
        help="Info about bulb(s) state",
    )
    parser.add_option(
        "",
        "--getclock",
        action="store_true",
        dest="getclock",
        default=False,
        help="Get clock",
    )
    parser.add_option(
        "",
        "--setclock",
        action="store_true",
        dest="setclock",
        default=False,
        help="Set clock to same as current time on this computer",
    )
    parser.add_option(
        "-t",
        "--timers",
        action="store_true",
        dest="showtimers",
        default=False,
        help="Show timers",
    )
    parser.add_option(
        "-T",
        "--settimer",
        dest="settimer",
        metavar="NUM MODE SETTINGS",
        default=None,
        nargs=3,
        help="Set timer. "
        + "NUM: number of the timer (1-6). "
        + "MODE: inactive, poweroff, default, color, preset, or warmwhite. "
        + "SETTINGS: a string of settings including time, repeatdays or date, "
        + "and other mode specific settings.   Use --timerhelp for more details.",
    )

    parser.add_option(
        "--protocol",
        dest="protocol",
        default=None,
        metavar="PROTOCOL",
        help="Set the device protocol. Currently only supports LEDENET",
    )

    other_group.add_option(
        "-v",
        "--volatile",
        action="store_true",
        dest="volatile",
        default=False,
        help="Don't persist mode setting with hard power cycle (RGB and WW modes only).",
    )
    parser.add_option_group(other_group)

    parser.usage = "usage: %prog [-sS10cwdkpCiltThe] [addr1 [addr2 [addr3] ...]."
    (options, args) = parser.parse_args()

    if options.showexamples:
        showUsageExamples()
        sys.exit(0)

    if options.timerhelp:
        showTimerHelp()
        sys.exit(0)

    if options.listpresets:
        for c in range(
            PresetPattern.seven_color_cross_fade, PresetPattern.seven_color_jumping + 1
        ):
            print(f"{c:2} {PresetPattern.valtostr(c)}")
        sys.exit(0)

    if options.listcolors:
        for c in utils.get_color_names_list():  # type: ignore
            print(f"{c}, ")
        print("")
        sys.exit(0)

    if options.settimer:
        new_timer = processSetTimerArgs(parser, options.settimer)
        options.new_timer = new_timer
    else:
        options.new_timer = None

    mode_count = 0
    if options.color:
        mode_count += 1
    if options.ww:
        mode_count += 1
    if options.cw:
        mode_count += 1
    if options.cct:
        mode_count += 1
    if options.preset:
        mode_count += 1
    if options.custom:
        mode_count += 1
    if mode_count > 1:
        parser.error(
            "options --color, --*white, --preset, --CCT, and --custom are mutually exclusive"
        )

    if options.on and options.off:
        parser.error("options --on and --off are mutually exclusive")

    if options.custom:
        options.custom = processCustomArgs(parser, options.custom)

    if options.color:
        options.color = utils.color_object_to_tuple(options.color)
        if options.color is None:
            parser.error("bad color specification")

    if options.preset:
        if not PresetPattern.valid(options.preset[0]):
            parser.error("Preset code is not in range")

    # asking for timer info, implicitly gets the state
    if options.showtimers:
        options.info = True

    op_count = mode_count
    if options.on:
        op_count += 1
    if options.off:
        op_count += 1
    if options.info:
        op_count += 1
    if options.getclock:
        op_count += 1
    if options.setclock:
        op_count += 1
    if options.listpresets:
        op_count += 1
    if options.settimer:
        op_count += 1

    if (not options.scan or options.scanresults) and (op_count == 0):
        parser.error("An operation must be specified")

    # if we're not scanning, IP addresses must be specified as positional args
    if not options.scan and not options.scanresults and not options.listpresets:
        if len(args) == 0:
            parser.error(
                "You must specify at least one IP address as an argument, or use scan results"
            )

    return (options, args)


# -------------------------------------------
def main() -> None:  # noqa: C901

    (options, args) = parseArgs()

    if options.scan:
        scanner = BulbScanner()
        scanner.scan(timeout=2)
        bulb_info_list = scanner.getBulbInfo()
        # we have a list of buld info dicts
        addrs = []
        if options.scanresults and len(bulb_info_list) > 0:
            for b in bulb_info_list:
                addrs.append(b["ipaddr"])
        else:
            print(f"{len(bulb_info_list)} bulbs found")
            for b in bulb_info_list:
                print("  {} {}".format(b["id"], b["ipaddr"]))
            sys.exit(0)

    else:
        addrs = args
        bulb_info_list = []
        for addr in args:
            bulb_info_list.append(FluxLEDDiscovery({"ipaddr": addr, "id": "Unknown ID"}))  # type: ignore

    # now we have our bulb list, perform same operation on all of them
    for info in bulb_info_list:
        try:
            bulb = WifiLedBulb(info["ipaddr"])
        except Exception as e:
            print("Unable to connect to bulb at [{}]: {}".format(info["ipaddr"], e))
            continue

        if options.getclock:
            print("{} [{}] {}".format(info["id"], info["ipaddr"], bulb.getClock()))

        if options.setclock:
            bulb.setClock()

        if options.protocol:
            bulb.setProtocol(options.protocol)

        if options.ww is not None:
            if options.ww > 100:
                print("Input can not be higher than 100%")
            else:
                print(f"Setting warm white mode, level: {options.ww}%")
                bulb.setWarmWhite(options.ww, not options.volatile)

        if options.cw is not None:
            if options.cw > 100:
                print("Input can not be higher than 100%")
            else:
                print(f"Setting cold white mode, level: {options.cw}%")
                bulb.setColdWhite(options.cw, not options.volatile)

        if options.cct is not None:
            if options.cct[1] > 100:
                print("Brightness can not be higher than 100%")
            elif options.cct[0] < 2700 or options.cct[0] > 6500:
                print("Color Temp must be between 2700 and 6500")
            else:
                print(
                    "Setting LED temperature {}K and brightness: {}%".format(
                        options.cct[0], options.cct[1]
                    )
                )
                bulb.setWhiteTemperature(
                    options.cct[0], options.cct[1] * 2.55, not options.volatile
                )

        if options.color is not None:
            print(
                f"Setting color RGB:{options.color}",
            )
            name = utils.color_tuple_to_string(options.color)
            if name is None:
                print()
            else:
                print(f"[{name}]")
            if any(i < 0 or i > 255 for i in options.color):
                print("Invalid value received must be between 0-255")
            elif len(options.color) == 3:
                bulb.setRgb(
                    options.color[0],
                    options.color[1],
                    options.color[2],
                    not options.volatile,
                )
            elif len(options.color) == 4:
                bulb.setRgbw(
                    options.color[0],
                    options.color[1],
                    options.color[2],
                    options.color[3],
                    not options.volatile,
                )
            elif len(options.color) == 5:
                bulb.setRgbw(
                    options.color[0],
                    options.color[1],
                    options.color[2],
                    options.color[3],
                    not options.volatile,
                    None,
                    options.color[4],
                )

        elif options.custom is not None:
            bulb.setCustomPattern(
                options.custom[2], options.custom[1], options.custom[0]
            )
            print(
                "Setting custom pattern: {}, Speed={}%, {}".format(
                    options.custom[0], options.custom[1], options.custom[2]
                )
            )

        elif options.preset is not None:
            print(
                "Setting preset pattern: {}, Speed={}%".format(
                    PresetPattern.valtostr(options.preset[0]), options.preset[1]
                )
            )
            bulb.setPresetPattern(options.preset[0], options.preset[1])

        if options.on:
            print(f"Turning on bulb at {bulb.ipaddr}")
            bulb.turnOn()
        elif options.off:
            print(f"Turning off bulb at {bulb.ipaddr}")
            bulb.turnOff()

        if options.info:
            bulb.update_state()
            print("{} [{}] {}".format(info["id"], info["ipaddr"], bulb))

        if options.settimer:
            timers = bulb.getTimers()
            num = int(options.settimer[0])
            print(f"New Timer ---- #{num}: {options.new_timer}")
            if options.new_timer.isExpired():
                print("[timer is already expired, will be deactivated]")
            timers[num - 1] = options.new_timer
            bulb.sendTimers(timers)

        if options.showtimers:
            timers = bulb.getTimers()
            num = 0
            for t in timers:
                num += 1
                print(f"  Timer #{num}: {t}")
            print("")

    sys.exit(0)


if __name__ == "__main__":
    main()



import asyncio
import contextlib
import logging
import time
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

from .scanner import BulbScanner, FluxLEDDiscovery

_LOGGER = logging.getLogger(__name__)


MAX_UPDATES_WITHOUT_RESPONSE = 4


class LEDENETDiscovery(asyncio.DatagramProtocol):
    def __init__(
        self,
        destination: Tuple[str, int],
        on_response: Callable[[bytes, Tuple[str, int]], None],
    ) -> None:
        """Init the discovery protocol."""
        self.transport = None
        self.destination = destination
        self.on_response = on_response

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """Trigger on_response."""
        self.on_response(data, addr)

    def error_received(self, ex: Optional[Exception]) -> None:
        """Handle error."""
        _LOGGER.error("LEDENETDiscovery error: %s", ex)

    def connection_lost(self, ex: Optional[Exception]) -> None:
        """The connection is lost."""


class AIOBulbScanner(BulbScanner):
    """A LEDENET discovery scanner."""

    def __init__(self) -> None:
        self.loop = asyncio.get_running_loop()
        super().__init__()

    async def _async_run_scan(
        self,
        transport: asyncio.DatagramTransport,
        destination: Tuple[str, int],
        timeout: int,
        found_all_future: "asyncio.Future[bool]",
    ) -> None:
        """Send the scans."""
        self.send_discovery_messages(transport, destination)
        quit_time = time.monotonic() + timeout
        time_out = timeout / self.BROADCAST_FREQUENCY
        while True:
            try:
                await asyncio.wait_for(
                    asyncio.shield(found_all_future), timeout=time_out
                )
            except asyncio.TimeoutError:
                pass
            else:
                return  # found_all
            time_out = min(
                quit_time - time.monotonic(), timeout / self.BROADCAST_FREQUENCY
            )
            if time_out <= 0:
                return
            # No response, send broadcast again in cast it got lost
            self.send_discovery_messages(transport, destination)

    async def async_scan(
        self, timeout: int = 10, address: Optional[str] = None
    ) -> List[FluxLEDDiscovery]:
        """Discover LEDENET."""
        sock = self._create_socket()
        destination = self._destination_from_address(address)
        found_all_future: "asyncio.Future[bool]" = asyncio.Future()
        response_list: Dict[str, Dict[str, Any]] = {}

        def _on_response(data: bytes, addr: Tuple[str, int]) -> None:
            _LOGGER.debug("discover: %s <= %s", addr, data)
            if self._process_response(data, addr, address, response_list):
                with contextlib.suppress(asyncio.InvalidStateError):
                    found_all_future.set_result(True)

        transport_proto = await self.loop.create_datagram_endpoint(
            lambda: LEDENETDiscovery(
                destination=destination,
                on_response=_on_response,
            ),
            sock=sock,
        )
        transport = cast(asyncio.DatagramTransport, transport_proto[0])
        try:
            await self._async_run_scan(
                transport, destination, timeout, found_all_future
            )
        finally:
            transport.close()

        self.found_bulbs = self._found_bulbs(response_list)
        return list(self.found_bulbs)
